/****   Copyright (c) 2001  Alexandros V. Gerbessiotis
        Read copyright notice available in file  "copy"
 ****/

#ifdef   MPILIB
MPI_Comm comm;
MPI_Win  win;
MPI_Win  win_max_ind;
MPI_Win  win_max_temp;
MPI_Win  win_column;
MPI_Win  win_row_tmp;
MPI_Win  win_row;
MPI_Win  win_collect_ind;
MPI_Win  win_collect_max;
int mpi_pid = -1;
int mpi_nprocs = -1;
double mpi_bsp_l = 300.0;
double mpi_bsp_g = 0.5;
double mpi_basetime=0.0;
#endif
