#include <memory.h>
#include <math.h>
#include "ai.h"

#define	SQRT(X)	((int)ceil(sqrt((double) (X))))
#define	SQRT_F(X)	((int)floor(sqrt((double) (X))))
#define CEIL(X,Y)       (((X) != 0)?(1+ (X-1)/Y):0)  
#define FLOOR(X,Y)      ((X)/(Y))  
#define LEFT(X,Y)       ((X)-(X/Y)*(Y))
#define POWER(X,Y)      pow((double)(X),(double)(Y))
#define EXP(X)          exp((double)(X))

#define MAX(X,Y)        ((X)>= (Y)?(X):(Y))

#define MIN(X,Y)        ((X)<= (Y)?(X):(Y))
